from abc import ABC, abstractmethod
import copy
from ..PlayerStructs import *
import json

class PlayerDataTrimAlg(ABC):

	def __init__(self):
		pass

	@abstractmethod
	def trimmedList(self, pastModelIncs):
		pass


# ---------------------- KNNRegression stuff ---------------------------
class AgeSortPlayerDataTrimAlg(PlayerDataTrimAlg):

	def __init__(self, maxNumModelElements):
		super().__init__()
		self.maxNumModelElements = maxNumModelElements

	def creationTimeSort(self, elem):
		return elem.creationTime

	def trimmedList(self, pastModelIncs):

		if(len(pastModelIncs) <= self.maxNumModelElements):
			return [pastModelIncs, []]

		pastModelIncsSorted = sorted(pastModelIncs, key=self.creationTimeSort)
		removedI = pastModelIncs.index(pastModelIncsSorted[0])
		pastModelIncs.pop(removedI)
		return [pastModelIncs, [removedI]]


class QualitySortPlayerDataTrimAlg(PlayerDataTrimAlg):

	def __init__(self, maxNumModelElements, qualityWeights = None, accStateResidue = None):
		super().__init__()
		self.maxNumModelElements = maxNumModelElements
		self.qualityWeights = PlayerCharacteristics(ability = 0.5, engagement = 0.5) if qualityWeights==None else qualityWeights
		self.accStateResidue = False if accStateResidue == None else accStateResidue

	def considerStateResidue(self, accStateResidue):
		self.accStateResidue = accStateResidue

	def stateTypeFilter(self, element):
		return element.stateType == 0

	def qSort(self, elem):
		return elem.quality

	def calcQuality(self, state):
		total = self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement
		return total

	def trimmedList(self, pastModelIncs):

		for modelInc in pastModelIncs:
			if(modelInc.quality == -1):
				modelInc.quality = self.calcQuality(modelInc)
				if(self.accStateResidue):
					modelInc.quality += modelInc.stateType


		if(len(pastModelIncs) <= self.maxNumModelElements):
			return [pastModelIncs, []]

		pastModelIncsSorted = sorted(pastModelIncs, key=self.qSort)
		removedI = pastModelIncs.index(pastModelIncsSorted[0])
		pastModelIncs.pop(removedI)
		return [pastModelIncs, [removedI]]


class ProximitySortPlayerDataTrimAlg(PlayerDataTrimAlg):

	def __init__(self, maxNumModelElements, epsilon = None, accStateResidue = None):
		super().__init__()
		self.maxNumModelElements = maxNumModelElements
		self.epsilon = 0.01 if epsilon == None else epsilon
		self.accStateResidue = False if accStateResidue == None else accStateResidue

	def considerStateResidue(self, accStateResidue):
		self.accStateResidue = accStateResidue
	
	def proximitySort(self, elem):
		return elem.quality

	def creationTimeSort(self, elem):
		return elem.creationTime

	def trimmedList(self, pastModelIncs):

		if(len(pastModelIncs) <= self.maxNumModelElements):
			return [pastModelIncs, []]

		pastModelIncsSortedAge = sorted(pastModelIncs, key=self.creationTimeSort)
		lastDataPoint = pastModelIncsSortedAge[-1]
		for modelInc in pastModelIncs:
			modelInc.quality = lastDataPoint.profile.sqrDistanceBetween(modelInc.profile)
			if(self.accStateResidue):
				modelInc.quality += modelInc.stateType

		# check if there is already a close point
		pastModelIncsSorted = sorted(pastModelIncs, key=self.proximitySort)
		pastModelIncsSorted.remove(lastDataPoint) #remove the point to be tested
		removedI = None
		closestPoint = pastModelIncsSorted[0]
		
		# print(json.dumps(closestPoint, default=lambda o: [o.__dict__["quality"],o.__dict__["stateType"],o.__dict__["creationTime"]], sort_keys=True))

		if (self.accStateResidue and closestPoint.stateType == 0) or closestPoint.quality > (self.epsilon + closestPoint.stateType):
			removedI = pastModelIncs.index(closestPoint)
			pastModelIncs.pop(removedI)
		else:
			removedI = pastModelIncs.index(lastDataPoint)
			pastModelIncs.pop(removedI)
		return [pastModelIncs, [removedI]]



import math
import copy
import random
import traceback


class InteractionsProfile(object):

	def __init__(self, dimensions = None):
		# if(dimensions != None):
		# 	if(type(dimensions)== dict):
		# 		self.dimensions = dimensions
		# 	elif(type(dimensions)== list):
		# 		self.unflatten(dimensions)
		# else:
		# 	self.dimensions = {}

		self.dimensions = {} if dimensions == None else dimensions
		self.dimensionality = len(self.dimensions)
		# self.normalize()

	def reset(self):
		for key in self.dimensions:
			self.dimensions[key] = 0
		return self

	def init(self):
		return self.reset()#.normalize()

	def generateCopy(self):
		keys = list(self.dimensions.keys())
		newVar = type(self)(copy.copy(self.dimensions))
		for key in keys:
			newVar.dimensions[key] = self.dimensions[key]
		return newVar

	def normalize(self):
		return self.normalization(self)

	def normalized(self):
		clone = self.generateCopy() 
		return self.normalization(clone)
	
	def normalization(self, profile):
		if(len(profile.dimensions)>1):
			total = 0
			for key in profile.dimensions:
				total += profile.dimensions[key]
			if(total==0):
				for key in profile.dimensions:
					profile.dimensions[key] = 1/len(profile.dimensions)
			else:
				for key in profile.dimensions:
					profile.dimensions[key] = profile.dimensions[key]/total
		return profile



	def randomize(self):
		return self.randomization(self)

	def randomized(self):
		clone = self.generateCopy() 
		return self.randomization(clone)

	def randomization(self, profile):
		profile.reset()
		for key in profile.dimensions:
			profile.dimensions[key] = random.uniform(0.0, 1.0)
		# profile.normalize()
		return profile


	def sqrDistanceBetween(self, profileToTest):
		cost = self.generateCopy()
		cost.reset()
		if(len(cost.dimensions) != len(profileToTest.dimensions)):
			traceback.print_stack()
			print(cost.dimensions)
			print(profileToTest.dimensions)
			raise Exception("[ERROR] Could not compute distance between profiles in different sized spaces. Execution aborted.")

		for key in cost.dimensions:
			cost.dimensions[key] = abs(self.dimensions[key] - profileToTest.dimensions[key])

		total = 0
		for key in cost.dimensions:
			cost.dimensions[key] = pow(cost.dimensions[key], 2)
			total += cost.dimensions[key]

		return total


	def distanceBetween(self, profileToTest):
		numDims = len(profileToTest.dimensions)
		return self.sqrDistanceBetween(profileToTest)**(1/float(numDims)) 


	def flattened(self):
		return [dim for dim in self.dimensions.values()]



	def unflattenFunc(self, profile, array):
		i = 0
		for key in profile.dimensions.keys():
			profile.dimensions[key] = array[i]
			i += 1
		return profile


	def unflatten(self, array):
		return self.unflattenFunc(self, array)

	def unflattened(self, array):
		clone = self.generateCopy() 
		return self.unflattenFunc(clone, array)


from abc import ABC, abstractmethod
from ..PlayerStructs import *

class TaskModelBridge(ABC):

	@abstractmethod
	def getAllTaskIds(self):
		pass

	@abstractmethod
	def getTaskInteractionsProfile(self, taskId):
		pass

	@abstractmethod
	def getMinTaskRequiredAbility(self, taskId):
		pass

	@abstractmethod
	def getMinTaskDuration(self, taskId):
		pass

	@abstractmethod
	def getTaskDifficultyWeight(self, taskId):
		pass

	@abstractmethod
	def getTaskProfileWeight(self, taskId):
		pass

	@abstractmethod
	def getTaskDiversityWeight(self, taskId):
		pass

	@abstractmethod
	def getTaskInitDate(self, taskId):
		pass

	@abstractmethod
	def getTaskFinalDate(self, taskId):
		pass


import math
from .AlgDefStructs.RegressionAlg import *
from .AlgDefStructs.ConfigsGenAlg import *
from .AlgDefStructs.PreferencesEstAlg import *

from .ModelBridge.PlayerModelBridge import PlayerModelBridge 
from .ModelBridge.TaskModelBridge import TaskModelBridge 


class Adaptation(object):

	def init(self, \
		playerModelBridge, \
		taskModelBridge, \
		name, \
		configsGenAlg):

		self.initialized = True
		self.playerIds = []
		self.taskIds = []
		self.name = name

		# self.numTasksPerGroup = numTasksPerGroup;
		self.configsGenAlg = configsGenAlg
		self.playerModelBridge = playerModelBridge
		self.taskModelBridge = taskModelBridge

		self.configsGenAlg.init()	


	# instrumentation
	def getConfigsGenAlgCompPerc(self):
		return self.configsGenAlg.getCompPercentage()

	def iterate(self):
		if not self.initialized:
			raise AssertionError('Adaptation not Initialized! Core not executed.') 
			return
		
		self.playerIds = self.playerModelBridge.getAllPlayerIds()
		self.taskIds = self.taskModelBridge.getAllTaskIds()

		if len(self.playerIds) < self.configsGenAlg.minNumberOfPlayersPerGroup:
			raise ValueError('Not enough players to form a group.') 
			return
		

		# print(json.dumps(self.playerModelBridge.getPlayerStatesDataFrame(0).states, default=lambda o: [o.__dict__["quality"],o.__dict__["stateType"],o.__dict__["creationTime"]], sort_keys=True))
		# print("\n\n")
		adaptedConfig = self.configsGenAlg.organize()

		adaptedGroups = adaptedConfig["groups"]
		adaptedProfiles = adaptedConfig["profiles"]
		adaptedAvgCharacteristics = adaptedConfig["avgCharacteristics"]
		adaptedConfig["tasks"] = []

		for groupIndex in range(len(adaptedGroups)):
			currGroup = adaptedGroups[groupIndex]
			groupProfile = adaptedProfiles[groupIndex]
			avgState = adaptedAvgCharacteristics[groupIndex]

			adaptedTaskId = self.selectTask(self.taskIds, groupProfile, avgState)
			for playerId in currGroup:

				currState = self.playerModelBridge.getPlayerCurrState(playerId)
				currState.profile = groupProfile	
				self.playerModelBridge.setPlayerTasks(playerId, [adaptedTaskId])
				self.playerModelBridge.setPlayerCharacteristics(playerId, currState.characteristics)
				self.playerModelBridge.setPlayerProfile(playerId, currState.profile)
				self.playerModelBridge.setPlayerGroup(playerId, currGroup)
				
			adaptedConfig["tasks"].append(adaptedTaskId)

			
		# totalFitness = 0.0
		# for groupI in range(len(adaptedGroups)):
		# 	group = adaptedGroups[groupI]
		# 	profile = adaptedProfiles[groupI]
		# 	for playerId in group:
		# 		predictedIncreases = self.configsGenAlg.regAlg.predict(profile, playerId)
		# 		totalFitness += (0.5* predictedIncreases.characteristics.ability + \
		# 						0.5* predictedIncreases.characteristics.engagement)
		
		# totalFitness = totalFitness + 1.0 #helps selection (otherwise Pchoice would always be 0)
		# print(totalFitness, end="\n")




		return adaptedConfig

	def selectTask(self,
		possibleTaskIds,
		bestConfigProfile,
		avgState):
		lowestCost = math.inf
		bestTaskId = -1 #if no tasks are available 

		for i in range(len(possibleTaskIds)):
			currTaskId = possibleTaskIds[i]

			cost = abs(bestConfigProfile.sqrDistanceBetween(self.taskModelBridge.getTaskInteractionsProfile(currTaskId)) * self.taskModelBridge.getTaskProfileWeight(currTaskId))
			cost += abs(avgState.ability - self.taskModelBridge.getMinTaskRequiredAbility(currTaskId) * self.taskModelBridge.getTaskDifficultyWeight(currTaskId))

			if cost < lowestCost:
				lowestCost = cost
				bestTaskId = currTaskId
				
		return bestTaskId





	# Bootstrap
	def simulateReaction(self, playerId):
		currState = self.playerModelBridge.getPlayerCurrState(playerId)
		newState = self.calcReaction(state = currState, playerId = playerId)

		increases = PlayerState(stateType = newState.stateType)
		increases.profile = currState.profile
		increases.characteristics = PlayerCharacteristics(ability=(newState.characteristics.ability - currState.characteristics.ability), engagement=newState.characteristics.engagement)
		self.playerModelBridge.setAndSavePlayerStateToDataFrame(playerId, increases, newState)	
		return increases

	def calcReaction(self, state, playerId):
		preferences = self.playerModelBridge.getPlayerRealPreferences(playerId)
		numDims = len(preferences.dimensions)
		newState = PlayerState(
			stateType = 0, 
			characteristics = PlayerCharacteristics(
				ability=state.characteristics.ability, 
				engagement=state.characteristics.engagement
				), 
			profile=state.profile)
		newState.characteristics.engagement = 1 - (preferences.distanceBetween(state.profile) / math.sqrt(numDims))  #between 0 and 1
		if newState.characteristics.engagement>1:
			raise ValueError('Something went wrong. Engagement is > 1.') 
		abilityIncreaseSim = (newState.characteristics.engagement*self.playerModelBridge.getBaseLearningRate(playerId))
		newState.characteristics.ability = newState.characteristics.ability + abilityIncreaseSim
		return newState

	def bootstrap(self, numBootstrapIterations):
		if(numBootstrapIterations <= 0):
			raise ValueError('Number of bootstrap iterations must be higher than 0 for this method to be called.') 
			return

		numPlayers = len(self.playerModelBridge.getAllPlayerIds())
		i = 0
		while(i < numBootstrapIterations):
			print("Performming step ("+str(i)+" of "+str(numBootstrapIterations)+") of the bootstrap phase of \""+str(self.name)+"\"...                                                             ", end="\r")
			self.iterate()
			for x in range(numPlayers):
				increases = self.simulateReaction(playerId=x)	
			i+=1


		self.configsGenAlg.reset()



import copy
import json
import numpy
import pandas as pd

from abc import ABC, abstractmethod

from GIMMECore.ModelBridge.TaskModelBridge import TaskModelBridge
from ..PlayerStructs import *

from sklearn import linear_model, neighbors

class RegressionAlg(ABC):

	def __init__(self, playerModelBridge):
		self.playerModelBridge = playerModelBridge

		self.completionPerc = 0.0

	@abstractmethod
	def predict(self, profile, playerId):
		pass

	@abstractmethod
	def isTabular(self):
		pass

	# If returns true, must implement a groupPredict() method
	def isGroupPredict(self):  
		return False
	
	def groupPredict(self, groupIds):
		pass
	
	# instrumentation
	def getCompPercentage(self):
		return self.completionPerc


# ---------------------- Personality Diversity ---------------------------
class DiversityValueAlg(RegressionAlg):
	#  Consider the task preferences of students in addition to team diversity. People with the same personality can still have different preferences
	#  Diversity weight is the value determined by the teacher (0 = aligned, 1 = diverse)
	def __init__(self, playerModelBridge, diversityWeight):
		super().__init__(playerModelBridge)
		self.diversityWeight = diversityWeight

	def predict(self, profile, playerId):
		return 0
	
	def isTabular(self):
		return False
	
	def isGroupPredict(self):
		return True
	
	def getPersonalitiesListFromPlayerIds(self, groupIds):
		personalities = []  # list of PlayerPersonality objects

		for playerId in groupIds:
			personality = self.playerModelBridge.getPlayerPersonality(playerId)
			if personality:
				personalities.append(personality)

		return personalities
	
	def getTeamPersonalityDiveristy(self, personalities):
		if len(personalities) <= 0:
			return -1
		
		diversity = -1

		if isinstance(personalities[0], PersonalityMBTI):
			diversity = PersonalityMBTI.getTeamPersonalityDiversity(personalities)

		return diversity


	def groupPredict(self, groupIds):
		personalities = self.getPersonalitiesListFromPlayerIds(groupIds)  # list of PlayerPersonality objects
		diversity = self.getTeamPersonalityDiveristy(personalities)

		# inverse of distance squared
		# lower distance = higher quality
		distance = abs(diversity - self.diversityWeight)
		
		if distance == 0.0:
			return 1.0
		
		return 1.0 / (distance * distance)


# class DiversityLogarithmicCentroidDistance(RegressionAlg):
# 	def __init__(self, playerModelBridge):
# 		super().__init__(playerModelBridge)

# ---------------------- Regression Based Characteristic Functions ---------------------------
class RegCoalitionValueAlg(RegressionAlg):
	def __init__(self, playerModelBridge, qualityWeights):
		super().__init__(playerModelBridge)
		self.qualityWeights = PlayerCharacteristics(ability = 0.5, engagement = 0.5) if qualityWeights == None else qualityWeights

	def isTabular(self):
		return False


# ---------------------- KNNRegression ---------------------------
class KNNRegression(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, numberOfNNs, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)
		self.numberOfNNs = numberOfNNs


	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement

	def distSort(self, elem):
		return elem.dist

	def creationTimeSort(self, elem):
		return elem.creationTime

	def predict(self, profile, playerId):
		# import time
		# startTime = time.time()

		pastModelIncs = self.playerModelBridge.getPlayerStatesDataFrame(playerId).getAllStates().copy()
		pastModelIncsSize = len(pastModelIncs)

		predictedState = PlayerState(profile = profile, characteristics = PlayerCharacteristics())

		for modelInc in pastModelIncs:
			modelInc.dist = profile.sqrDistanceBetween(modelInc.profile)

		pastModelIncs = sorted(pastModelIncs, key=self.distSort)

		numberOfIterations = min(self.numberOfNNs, len(pastModelIncs))
		pastModelIncs = pastModelIncs[:numberOfIterations]

		triangularNumberOfIt = sum(range(numberOfIterations + 1))
		for i in range(numberOfIterations):

			self.completionPerc = i/ numberOfIterations

			currState = pastModelIncs[i]
			pastCharacteristics = currState.characteristics
			ratio = (numberOfIterations - i)/triangularNumberOfIt

			predictedState.characteristics.ability += pastCharacteristics.ability * ratio
			predictedState.characteristics.engagement += pastCharacteristics.engagement * ratio

		# executionTime = (time.time() - startTime)
		# print('Execution time in seconds: ' + str(executionTime))
		self.state = predictedState

		return self.calcQuality(predictedState)

# ---------------------- KNNRegressionSKLearn ---------------------------
class KNNRegressionSKLearn(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, numberOfNNs, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)
		self.numberOfNNs = numberOfNNs

	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement

	def predict(self, profile, playerId):
		# import time
		# startTime = time.time()

		pastModelIncs = self.playerModelBridge.getPlayerStatesDataFrame(playerId).getAllStatesFlatten()

		lenPMI = len(pastModelIncs['profiles'])

		numberOfNNs = self.numberOfNNs
		if(lenPMI < self.numberOfNNs):
			if(lenPMI==0):
				return PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = 0.5, engagement = 0.5))
			numberOfNNs = lenPMI

		profData = profile.flattened()
		prevProfs = pastModelIncs['profiles']


		self.regrAb = neighbors.KNeighborsRegressor(numberOfNNs, weights="distance")
		self.regrAb.fit(prevProfs, pastModelIncs['abilities'])
		predAbilityInc = self.regrAb.predict([profData])[0]

		self.regrEng = neighbors.KNeighborsRegressor(numberOfNNs, weights="distance")
		self.regrEng.fit(prevProfs, pastModelIncs['engagements'])
		predEngagement = self.regrEng.predict([profData])[0]

		predState = PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = predAbilityInc, engagement = predEngagement))


		self.completionPerc = 1.0

		# executionTime = (time.time() - startTime)
		# print('Execution time in seconds: ' + str(executionTime))
		self.state = predState
		return self.calcQuality(predState)

# ---------------------- LinearRegressionSKLearn ---------------------------
class LinearRegressionSKLearn(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)


	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement


	def predict(self, profile, playerId):

		pastModelIncs = self.playerModelBridge.getPlayerStatesDataFrame(playerId).getAllStatesFlatten()

		if(len(pastModelIncs['profiles'])==0):
			return PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = 0.5, engagement = 0.5))

		profData = profile.flattened()

		prevProfs = pastModelIncs['profiles']

		regr = linear_model.LinearRegression()
		regr.fit(prevProfs, pastModelIncs['abilities'])
		predAbilityInc = regr.predict([profData])[0]

		regr.fit(prevProfs, pastModelIncs['engagements'])
		predEngagement = regr.predict([profData])[0]

		predState = PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = predAbilityInc, engagement = predEngagement))

		self.completionPerc = 1.0
		self.state = predState
		return self.calcQuality(predState)

# ---------------------- SVMRegressionSKLearn ---------------------------
class SVMRegressionSKLearn(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)


	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement


	def predict(self, profile, playerId):

		pastModelIncs = self.playerModelBridge.getPlayerStatesDataFrame(playerId).getAllStatesFlatten()

		if(len(pastModelIncs['profiles'])==0):
			return PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = 0.5, engagement = 0.5))

		profData = profile.flattened()

		prevProfs = pastModelIncs['profiles']

		regr = svm.SVR()
		regr.fit(prevProfs, pastModelIncs['abilities'])
		predAbility = regr.predict([profData])[0]

		regr.fit(prevProfs, pastModelIncs['engagements'])
		predEngagement = regr.predict([profData])[0]

		predState = PlayerState(profile = profile, characteristics = PlayerCharacteristics(ability = predAbility, engagement = predEngagement))

		self.completionPerc = 1.0
		self.state = predState
		return self.calcQuality(predState)

# ---------------------- DecisionTreesRegression ---------------------------
class DecisionTreesRegression(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)

	def predict(self, profile, playerId):
		pass


# ---------------------- NeuralNetworkRegression ---------------------------
class NeuralNetworkRegression(RegCoalitionValueAlg):

	def __init__(self, playerModelBridge, qualityWeights = None):
		super().__init__(playerModelBridge, qualityWeights)

	def predict(self, profile, playerId):
		pass


# ---------------------- Tabular Characteristic Functions -------------------------------------
class TabularCoalitionValueAlg(RegressionAlg):
	def __init__(self, playerModelBridge):
		super().__init__(playerModelBridge)
		self.playerPrefEstimates = {}


	def isTabular(self):
		return True

	def getPlayerPreferencesEstimations(self):
	 	for player in self.playerIds:
	 		self.playerPrefEstimates[player] = self.playerModelBridge.getPlayerPreferencesEst(player)

# ---------------------- Tabular Agent Synergy Method -------------------------------------
class TabularAgentSynergies(TabularCoalitionValueAlg):

	def __init__(self, playerModelBridge, taskModelBridge, syntergyTablePath):
		super().__init__(playerModelBridge)


		self.taskModelBridge = taskModelBridge
		tempTable = pd.read_csv(syntergyTablePath, sep=",", dtype={'agent_1': object, 'agent_2': object})
		synergyTable = tempTable.pivot_table(values='synergy', index='agent_1', columns='agent_2')

		self.synergyMatrix = synergyTable.to_numpy()
		self.synergyMatrix[numpy.isnan(self.synergyMatrix)] = 0
		self.synergyMatrix = self.symmetrize(self.synergyMatrix)

		self.playerIds = self.playerModelBridge.getAllPlayerIds()

		# tempTable = pd.read_csv('taskTable.txt', sep=',', dtype={'task': object, 'agent': object})
		# taskTable = tempTable.pivot_table(values='synergy', index='task', columns='agent')

		# self.taskMatrix = taskTable.to_numpy()
		# self.taskMatrix[numpy.isnan(self.taskMatrix)] = 0

	def symmetrize(self, table):
		return table + table.T - numpy.diag(table.diagonal())

	def predict(self, profile, playerId):
		firstPlayerPreferencesInBinary = ''
		for dim in profile.dimensions:
			firstPlayerPreferencesInBinary += str(round(profile.dimensions[dim]))

		if (self.playerPrefEstimates == {}):
			self.getPlayerPreferencesEstimations()

		secondPlayerPreferences = self.playerPrefEstimates[playerId]
		secondPlayerPreferenceInBinary = ''
		for dim in secondPlayerPreferences.dimensions:
			secondPlayerPreferenceInBinary += str(round(secondPlayerPreferences.dimensions[dim]))

		firstPlayerPreferencesIndex = int(firstPlayerPreferencesInBinary, 2)
		secondPlayerPreferencesIndex = int(secondPlayerPreferenceInBinary, 2)

		return self.synergyMatrix[firstPlayerPreferencesIndex][secondPlayerPreferencesIndex]


	# either this, or find here the best task
	def predictTasks(self, taskId, playerId):
		playerPreferences = self.playerPrefEstimates[playerId]
		playerPreferenceInBinary = ''
		for dim in playerPreferences.dimensions:
			playerPreferenceInBinary += str(round(playerPreferences.dimensions[dim]))

		taskProfile = self.taskModelBridge.getTaskInteractionsProfile(taskId)
		taskProfileInBinary = ''
		for dim in taskProfile.dimensions:
			taskProfileInBinary += str(round(taskProfile.dimensions[dim]))

		playerPreferenceIndex = int(playerPreferenceInBinary, 2)
		taskProfileIndex = int(taskProfileInBinary, 2)

		return self.taskMatrix[playerPreferenceIndex][taskProfileIndex]





from .Adaptation import Adaptation
from .AlgDefStructs.RegressionAlg import *
from .AlgDefStructs.ConfigsGenAlg import *
from .AlgDefStructs.PreferencesEstAlg import *
from .AlgDefStructs.PlayerDataTrimAlg import *

from .PlayerStructs import *
from .InteractionsProfile import InteractionsProfile


from .ModelBridge.PlayerModelBridge import PlayerModelBridge 
from .ModelBridge.TaskModelBridge import TaskModelBridge 


from abc import ABC, abstractmethod
from ..PlayerStructs import *

class PlayerModelBridge(ABC):
	@abstractmethod
	def resetPlayer(self, playerId):
		pass

	@abstractmethod
	def getAllPlayerIds(self):
		pass

	@abstractmethod
	def getPlayerName(self, playerId):
		pass


	@abstractmethod
	def getPlayerCurrState(self,  playerId):
		pass

	@abstractmethod
	def getPlayerPersonality(self,  playerId):
		pass

	@abstractmethod
	def getPlayerCurrCharacteristics(self, playerId):
		pass


	@abstractmethod
	def getPlayerStatesDataFrame(self, playerId):
		pass

	@abstractmethod
	def getPlayerPreferencesEst(self, playerId):
		pass

	@abstractmethod
	def setPlayerPreferencesEst(self, playerId, preferences):
		pass

	@abstractmethod
	def setAndSavePlayerStateToDataFrame(self, playerId, increases, newState):
		pass

	
	@abstractmethod
	def setPlayerCharacteristics(self, playerId, characteristics):
		pass

	@abstractmethod		
	def setPlayerGroup(self, playerId, group):
		pass

	@abstractmethod		
	def setPlayerTasks(self, playerId, tasks):
		pass

	@abstractmethod		
	def setPlayerProfile(self, playerId, profile):
		pass


import math
#from sre_parse import State
import time
import copy
import json
from .InteractionsProfile import InteractionsProfile

from abc import ABC, abstractmethod


class PlayerCharacteristics(object):
	def __init__(self, ability=None, engagement=None, group_diversity=None):
		self.ability = 0 if ability == None else ability
		self.engagement = 0 if engagement == None else engagement
		self.group_diversity = 0 if group_diversity == None else group_diversity

	# self._ability  = 0 if ability==None else min(ability, 20.0)
	# self._engagement = 0 if engagement==None else min(engagement, 1.0)

	def reset(self):
		self.ability = 0
		self.engagement = 0
		return self

# #constraints enforced via properties
# @property
# def ability(self):
# 	return self._ability

# @property
# def engagement(self):
# 	return self._engagement

# @ability.setter
# def ability(self, newValue):
# 	# print(newValue)
# 	self._ability = min(newValue, 20.0)
# 	# self._ability = newValue
# 	return self._ability

# @engagement.setter
# def engagement(self, newValue):
# 	self._engagement = min(newValue, 1.0)
# 	# self._engagement = newValue
# 	return self._engagement


class PlayerState(object):
	def __init__(self, stateType=None, profile=None, characteristics=None, dist=None, quality=None, group=None,
				 tasks=None):
		self.creationTime = time.time()

		self.stateType = 1 if stateType == None else stateType
		self.profile = InteractionsProfile() if profile == None else profile
		self.characteristics = PlayerCharacteristics() if characteristics == None else characteristics
		self.dist = -1 if dist == None else dist
		self.quality = -1 if quality == None else quality

		self.group = [] if group == None else group
		self.tasks = [] if tasks == None else tasks

	def reset(self):
		self.characteristics.reset()
		self.profile.reset()
		self.creationTime = time.time()
		self.stateType = 1
		self.quality = -1
		self.dist = -1

		self.group = []
		self.tasks = []
		return self
	

class PlayerPersonality(ABC):
	def __init__(self):
		self.maxDifferenceValue = 1

	@abstractmethod
	def getPersonalityString(self):
		pass

	@abstractmethod
	def getPairPersonalityDiversity(self, other):
		pass

	@abstractmethod
	def getTeamPersonalityDiversity(players):
		pass


class PersonalityMBTI(PlayerPersonality):
	def __init__(self):
		super()
			
	def __init__(self, letter1, letter2, letter3, letter4):
		super()
		self.setLetters(letter1, letter2, letter3, letter4)

	def setLetters(self, letter1, letter2, letter3, letter4):
		self.letter1 = letter1.upper()
		self.letter2 = letter2.upper()
		self.letter3 = letter3.upper()
		self.letter4 = letter4.upper()

	def getPersonalityString(self):
		return self.letter1 + self.letter2 + self.letter3 + self.letter4

	def getLettersList(self):
		return [self.letter1, self.letter2, self.letter3, self.letter4]

	# Determine the difference between 2 personalities. The value ranges from 0 (no difference) to 1 (max difference). 
	def getPairPersonalityDiversity(self, other):
		if not isinstance(other, PersonalityMBTI):
			raise Exception("[ERROR] Comparison between different personality models not allowed.")

		difference = 0
		otherLetters = other.getLettersList()
		selfLetters = self.getLettersList()

		for i in range(0, len(selfLetters)):
			difference += 0 if selfLetters[i] == otherLetters[i] else self.maxDifferenceValue / 4

		return difference

	# Determine the group personality difference. Using a formula proposed by Pieterse, Kourie and Sonnekus.
	# players is a list of PlayerPersonalties
	def getTeamPersonalityDiversity(players):
		# Difference is 0 if there's only one player
		if len(players) == 1:
			return 0

		letters_list = [[], [], [], []]

		# Populate letters_list with all the players' letters
		for player in players:
			for letters, letter in zip(letters_list, player.getLettersList()):
				letters.append(letter)
			
		difference = 0.0

		for letters in letters_list:
			letters.sort()
			length = len(letters)


			if (letters[0] != letters[1]) or (letters[length-2] != letters[length-1]):
				# The first/last two letters are different -> means all but one letters are the same (difference = 1)
				difference += 1.0
				continue
			elif length <= 3:
				# The first/last two letters are the same and len =< 3, means all the letters are the same (difference = 0)
				continue

			for letter in letters:
				# If not all the letters are the same, then difference = 2
				if letter != letters[1]:
					difference += 2.0
					break

			# Otherwise, all the letters are the same (difference = 0)

		# Max value for difference is 8. Divide by 8 in order to normalize it.
		return difference / 8.0


class PlayerStatesDataFrame(object):
	def __init__(self, interactionsProfileTemplate, trimAlg, states=None):
		self.interactionsProfileTemplate = interactionsProfileTemplate
		self.trimAlg = trimAlg

		self.states = [] if states == None else states

		# auxiliary stuff
		self.flatProfiles = []
		self.flatAbilities = []
		self.flatEngagements = []
		if states != None:
			for state in self.states:
				self.flatProfiles.append(state.profile.flattened())
				self.flatAbilities.append(state.characteristics.ability)
				self.flatEngagements.append(state.characteristics.engagement)

	def reset(self):
		self.states = []

		# auxiliary stuff
		self.flatProfiles = []
		self.flatAbilities = []
		self.flatEngagements = []

		return self

	def pushToDataFrame(self, playerState):
		self.states.append(playerState)

		# update tuple representation
		self.flatProfiles.append(playerState.profile.flattened())
		self.flatAbilities.append(playerState.characteristics.ability)
		self.flatEngagements.append(playerState.characteristics.engagement)

		# print(json.dumps(self.states, default=lambda o: o.__dict__, sort_keys=True, indent=2))
		# print("---------")
		# print(self.flatProfiles)
		# print(self.flatAbilities)
		# print(self.flatEngagements)

		trimmedListAndRemainder = self.trimAlg.trimmedList(self.states)
		trimmedList = trimmedListAndRemainder[0]
		remainderIndexes = trimmedListAndRemainder[1]

		# print(remainderIndexes)

		self.states = trimmedList

		# update tuple representation
		for i in remainderIndexes:
			self.flatProfiles.pop(i)
			self.flatAbilities.pop(i)
			self.flatEngagements.pop(i)

	def getAllStates(self):
		return self.states

	def getAllStatesFlatten(self):
		return {'profiles': self.flatProfiles, 'abilities': self.flatAbilities, 'engagements': self.flatEngagements}

	def getNumStates(self):
		return len(self.states)



import gc
import random
import math
import copy
from sys import breakpointhook
import numpy
import re
from abc import ABC, abstractmethod

import GIMMESolver as gs
from ctypes import *

from ..InteractionsProfile import InteractionsProfile 
from ..PlayerStructs import *
from ..AlgDefStructs.RegressionAlg import *


class ConfigsGenAlg(ABC):

	def __init__(self, 
		playerModelBridge,
		interactionsProfileTemplate, 
		taskModelBridge = None, 
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None,
		jointPlayerConstraints = "",
		separatedPlayerConstraints = ""):

		self.groupSizeFreqs = {}
		self.configSizeFreqs = {}

		self.jointPlayersConstraints = []
		self.separatedPlayersConstraints = []
		self.allConstraints = []
		
		minNumberOfPlayersPerGroup = 2 if minNumberOfPlayersPerGroup == None else minNumberOfPlayersPerGroup 
		maxNumberOfPlayersPerGroup = 5 if maxNumberOfPlayersPerGroup == None else maxNumberOfPlayersPerGroup

		if(minNumberOfPlayersPerGroup > maxNumberOfPlayersPerGroup):
			raise ValueError('The min number of players per group cannot be higher than the max!') 
		
		if preferredNumberOfPlayersPerGroup == None:
			self.maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup
			self.minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup
		else:
			self.maxNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup
			self.minNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup

		self.playerModelBridge = playerModelBridge
		self.taskModelBridge = taskModelBridge
		self.interactionsProfileTemplate = interactionsProfileTemplate
		
		jointPlayerConstraints = self.fromStringConstraintToList(jointPlayerConstraints)
		separatedPlayerConstraints = self.fromStringConstraintToList(separatedPlayerConstraints)
		
		for i in range(len(jointPlayerConstraints)):
			if jointPlayerConstraints[i] == ['']:
				continue
			self.addJointPlayersConstraints(jointPlayerConstraints[i])
			
		for i in range(len(separatedPlayerConstraints)):
			if separatedPlayerConstraints[i] == ['']:
				continue
			self.addSeparatedPlayersConstraints(separatedPlayerConstraints[i])

		self.completionPerc = 0.0


	def init(self):
		self.groupSizeFreqs = {}
		self.configSizeFreqs = {}
		return self

	def reset(self):
		return self.init()

	def randomConfigGenerator(self, playerIds, minNumGroups, maxNumGroups):
		
		returnedConfig = []
		playerJointRequirements = {}
		playerSeparatedRequirements = {}
		#listOfPlayersWithJointRequirements = []
		if self.jointPlayersConstraints != []:
			for id in playerIds:
				playerJointRequirements[str(id)] = []

			for constraint in self.jointPlayersConstraints:
				for i in range(len(constraint)):
					
					for j in range(len(constraint)):
						if constraint[i] == constraint[j]:
							continue

						playerJointRequirements[constraint[i]].append(constraint[j])


			for id in playerIds:
				for restrictedId in playerJointRequirements[id]:
					for restrictionOfRestrictedId in playerJointRequirements[restrictedId]:
						if restrictionOfRestrictedId not in playerJointRequirements[id] and restrictionOfRestrictedId != restrictedId:
							playerJointRequirements[id].append(restrictionOfRestrictedId)

		if self.separatedPlayersConstraints != []:
			for id in playerIds:
				playerSeparatedRequirements[str(id)] = []

			for constraint in self.separatedPlayersConstraints:
				for i in range(len(constraint)):
					
					for j in range(len(constraint)):
						if constraint[i] == constraint[j]:
							continue

						playerSeparatedRequirements[constraint[i]].append(constraint[j])

		if(len(playerIds) < self.minNumberOfPlayersPerGroup):
			print("number of players is lower than the minimum number of players per group!")
			return bestGroups
			
		# generate random config
		playersWithoutGroup = playerIds.copy()

		if(minNumGroups < maxNumGroups):
			numGroups = numpy.random.randint(minNumGroups, maxNumGroups)
		else: # players length is 1
			numGroups = maxNumGroups

		# generate min num players for each group
		playersWithoutGroupSize = len(playersWithoutGroup)
		# if (listOfPlayersWithJointRequirements != []):
		# 	playersWithoutGroup = listOfPlayersWithJointRequirements.copy()

		# playersWithoutGroupWithoutRestrictions = list(set(playersWithoutGroup) - set(listOfPlayersWithJointRequirements))
		for g in range(numGroups):
			currGroup = []
			currGroupSeparationRestrictions = []

			if(playersWithoutGroupSize < 1):
				break

			# add min number of players to the group
			for p in range(self.minNumberOfPlayersPerGroup):
				currPlayerIndex = random.randint(0, len(playersWithoutGroup) - 1)
				
				currPlayerID = playersWithoutGroup[currPlayerIndex]
				currGroup.append(currPlayerID)
				del playersWithoutGroup[currPlayerIndex]

			if ((playerSeparatedRequirements != {} or playerJointRequirements != {}) and len(playersWithoutGroup) > 0 ):
				self.verifyCoalitionValidity(currGroup, playerJointRequirements, playerSeparatedRequirements, playersWithoutGroup)
			returnedConfig.append(currGroup)
		
		# append the rest
		playersWithoutGroupSize = len(playersWithoutGroup)
		while playersWithoutGroupSize > 0:
			currPlayerIndex = 0;
			if (playersWithoutGroupSize > 1):
				currPlayerIndex = random.randint(0, playersWithoutGroupSize - 1)
			else:
				currPlayerIndex = 0
			currPlayerID = playersWithoutGroup[currPlayerIndex]

			groupsSize = len(returnedConfig)

			availableGroups = returnedConfig.copy()
			while (len(currGroup) > (self.maxNumberOfPlayersPerGroup - 1)):
				if(len(availableGroups) < 1):
					currGroup = random.choice(returnedConfig)
					break
				currGroup = random.choice(availableGroups)
				availableGroups.remove(currGroup)

			currGroup.append(currPlayerID)

			del playersWithoutGroup[currPlayerIndex]
			playersWithoutGroupSize = len(playersWithoutGroup)
		
		return returnedConfig

	def verifyCoalitionValidity(self, config, playerJointRequirements, playerSeparatedRequirements, playersWithoutGroup):
		for i in range(len(config)):
			if playerJointRequirements[config[i]] != []:
				playersNotInCoalition = []
				for player in playerJointRequirements[config[i]]:
					if player not in config:
						playersNotInCoalition.append(player)
				
				if playersNotInCoalition != []:
					
					for j in range(len(config)):
						if i != j and playersNotInCoalition[0] in playersWithoutGroup and config[j] not in playerJointRequirements[config[i]]:
							playersWithoutGroup.append(config[j])
							config[j] = playersNotInCoalition[0]
							playersWithoutGroup.remove(playersNotInCoalition[0])
							del playersNotInCoalition[0]

							if len(playersNotInCoalition) == 0:
								break
			
			if playerSeparatedRequirements[config[i]] != []:
				for player in playerSeparatedRequirements[config[i]]:
					if player in config:
						currPlayerIndex = random.randint(0, len(playersWithoutGroup) - 1)
						while playersWithoutGroup[currPlayerIndex] in playerSeparatedRequirements[config[i]]:
							currPlayerIndex = random.randint(0, len(playersWithoutGroup) - 1)

						config.remove(player)
						config.append(playersWithoutGroup[currPlayerIndex])
						del playersWithoutGroup[currPlayerIndex]

						playersWithoutGroup.append(player)
					
		return config


	def fromStringConstraintToList(self, constraints):
		constraints = constraints.split(';')

		for i in range(len(constraints)):
			constraints[i] = re.sub('[^A-Za-z0-9,_]+', '', constraints[i]).split(',')
		
		return constraints

	def addJointPlayersConstraints(self, players):
		self.jointPlayersConstraints.append(players)
		self.allConstraints.append({"players": players, "type": "JOIN"})

	def addSeparatedPlayersConstraints(self, players):
		self.separatedPlayersConstraints.append(players)
		self.allConstraints.append({"players": players, "type": "SEPARATE"})


	def resetPlayersConstraints(self):
		self.jointPlayersConstraints = []
		self.separatedPlayersConstraints = []
		self.allConstraints = []

	def getPlayerConstraints(self):
		return self.allConstraints

	@abstractmethod
	def organize(self):
		pass

	def updateMetrics(self, groups):

		# kind of sub-optimal, but guarantees encapsulation
		if(self.configSizeFreqs.get(len(groups))):
			self.configSizeFreqs[len(groups)]+=1
		else:
			self.configSizeFreqs[len(groups)]=1

		for group in groups:
			if(self.configSizeFreqs.get(len(group))):
				self.configSizeFreqs[len(group)]+=1
			else:
				self.configSizeFreqs[len(group)]=1


	def getCompPercentage(self):
		return self.completionPerc



class RandomConfigsGen(ConfigsGenAlg):

	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None,
		jointPlayerConstraints = "",
		separatedPlayerConstraints = ""):
		super().__init__(
			playerModelBridge = playerModelBridge,
			interactionsProfileTemplate = interactionsProfileTemplate, 
			preferredNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup, 
			minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup, 
			maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup,
			jointPlayerConstraints = jointPlayerConstraints,
			separatedPlayerConstraints = separatedPlayerConstraints)

	def organize(self):
		playerIds = self.playerModelBridge.getAllPlayerIds() 
		minNumGroups = math.ceil(len(playerIds) / self.maxNumberOfPlayersPerGroup)
		maxNumGroups = math.floor(len(playerIds) / self.minNumberOfPlayersPerGroup)
		
		newConfigProfiles = []
		newAvgCharacteristics = []
		newGroups = self.randomConfigGenerator(playerIds, minNumGroups, maxNumGroups)
		newConfigSize = len(newGroups)		
		# generate profiles
		for groupI in range(newConfigSize):
			group = newGroups[groupI]
			groupSize = len(group)


			profile = self.interactionsProfileTemplate.generateCopy().randomize()			
			# generate random profile
			# for currPlayer in group:
			# 	random = self.interactionsProfileTemplate.generateCopy().randomize()
			# 	for dim in profile.dimensions:
			# 		profile.dimensions[dim] += random.dimensions[dim] / groupSize
			# profile.normalize()
			newConfigProfiles.append(profile)


			currAvgCharacteristics = PlayerCharacteristics().reset()
			for currPlayer in group:
				currState = self.playerModelBridge.getPlayerCurrState(currPlayer)
				currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
				currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize			
			# currAvgCharacteristics.profile = profile

			diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
			personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(group)
			currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

			newAvgCharacteristics.append(currAvgCharacteristics)

			self.completionPerc = groupI/newConfigSize

		self.updateMetrics(newGroups)
		#print(newGroups)
		return {"groups": newGroups, "profiles": newConfigProfiles, "avgCharacteristics": newAvgCharacteristics}




class AnnealedPRSConfigsGen(ConfigsGenAlg):

	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg, 
		persEstAlg, 
		temperatureDecay, 
		numberOfConfigChoices = None, 
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None):

		super().__init__(
			playerModelBridge = playerModelBridge,
			interactionsProfileTemplate = interactionsProfileTemplate, 
			preferredNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup, 
			minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup, 
			maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup)

		self.regAlg = regAlg
		self.persEstAlg = persEstAlg
		self.numberOfConfigChoices = 100 if numberOfConfigChoices == None else numberOfConfigChoices

		self.temperature = 1.0
		self.temperatureDecay = temperatureDecay


	def init(self):
		super().init()
		self.temperature = 1.0

	def reset(self, temperature):
		super().reset()
		self.temperature = numpy.clip(temperature, 0, 1)

	def organize(self):
		playerIds = self.playerModelBridge.getAllPlayerIds() 
		minNumGroups = math.ceil(len(playerIds) / self.maxNumberOfPlayersPerGroup)
		maxNumGroups = math.floor(len(playerIds) / self.minNumberOfPlayersPerGroup)

		currMaxQuality = -float("inf")
		bestGroups = []
		bestConfigProfiles = []
		bestAvgCharacteristics = []


		# estimate preferences
		self.playerPrefEstimates = self.persEstAlg.updateEstimates()

		if (self.regAlg.isTabular()):
			self.regAlg.playerPrefEstimates = self.playerPrefEstimates

		playersCurrState = {}
		for player in self.playerIds:
			playersCurrState[player] = self.playerModelBridge.getPlayerCurrState(player)

		# generate several random groups, calculate their fitness and select the best one
		for i in range(self.numberOfConfigChoices):
			
			# generate several random groups
			newGroups = self.randomConfigGenerator(playerIds, minNumGroups, maxNumGroups)
			newConfigSize = len(newGroups)
			currQuality = 0.0
			newConfigProfiles = []
			newAvgCharacteristics = []

			# generate profiles
			for groupI in range(newConfigSize):
				group = newGroups[groupI]
				groupSize = len(group)

				# generate group profile as random or average of the preferences estimates
				profile = self.interactionsProfileTemplate.generateCopy().reset()

				if(random.uniform(0.0, 1.0) > self.temperature):
					for currPlayer in group:
						preferences = self.playerPrefEstimates[currPlayer]
						for dim in profile.dimensions:
							profile.dimensions[dim] += preferences.dimensions[dim] / groupSize
					# profile.normalize()
				else:
					profile = self.interactionsProfileTemplate.generateCopy().randomize()

				newConfigProfiles.append(profile)
				
				# calculate fitness and average state
				currAvgCharacteristics = PlayerCharacteristics()
				currAvgCharacteristics.reset()
				for i in range(len(group)):

					currState = playersCurrState[group[i]]
					currState.profile = profile

					currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
					currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize
				
					if (self.regAlg.isTabular()):
						firstPlayerPreferences = self.playerPrefEstimates[group[i]]
						for j in range(i+1, groupSize):
							currQuality += self.regAlg.predict(firstPlayerPreferences, group[j]) / math.comb(groupSize, 2)

					elif (not self.regAlg.isGroupPredict()):
						currQuality += self.regAlg.predict(profile, group[i])


				if (self.regAlg.isGroupPredict()):
					currQuality += self.regAlg.groupPredict(group)
		
				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(group)
				currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

				newAvgCharacteristics.append(currAvgCharacteristics)
			
			if (currQuality > currMaxQuality):
				bestGroups = newGroups
				bestConfigProfiles = newConfigProfiles
				bestAvgCharacteristics = newAvgCharacteristics
				currMaxQuality = currQuality


			self.completionPerc = i/self.numberOfConfigChoices

		if(self.temperature > 0.0):
			self.temperature -= self.temperatureDecay
		else:
			self.temperature = 1.0

		self.updateMetrics(bestGroups)

		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": bestAvgCharacteristics}




class PureRandomSearchConfigsGen(ConfigsGenAlg):

	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg, 
		persEstAlg, 
		numberOfConfigChoices = None, 
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None,
		jointPlayerConstraints = "",
		separatedPlayerConstraints = ""):
		
		super().__init__(
			playerModelBridge = playerModelBridge,
			interactionsProfileTemplate = interactionsProfileTemplate, 
			preferredNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup, 
			minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup, 
			maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup,
			jointPlayerConstraints = jointPlayerConstraints,
			separatedPlayerConstraints = separatedPlayerConstraints)

		self.regAlg = regAlg
		self.persEstAlg = persEstAlg
		self.numberOfConfigChoices = 100 if numberOfConfigChoices == None else numberOfConfigChoices
	
	def organize(self):
		playerIds = self.playerModelBridge.getAllPlayerIds() 
		minNumGroups = math.ceil(len(playerIds) / self.maxNumberOfPlayersPerGroup)
		maxNumGroups = math.floor(len(playerIds) / self.minNumberOfPlayersPerGroup)

		currMaxQuality = -float("inf")
		bestGroups = []
		bestConfigProfiles = []
		bestAvgCharacteristics = []


		# estimate preferences
		self.playerPrefEstimates = self.persEstAlg.updateEstimates()

		if (self.regAlg.isTabular()):
			self.regAlg.playerPrefEstimates = self.playerPrefEstimates

		# generate several random groups, calculate their fitness and select the best one
		for i in range(self.numberOfConfigChoices):
			
			# generate several random groups
			newGroups = self.randomConfigGenerator(playerIds, minNumGroups, maxNumGroups)
			newConfigSize = len(newGroups)
			currQuality = 0.0
			newConfigProfiles = []
			newAvgCharacteristics = []

			# generate profiles
			for groupI in range(newConfigSize):
				group = newGroups[groupI]
				groupSize = len(group)

				# generate profile as average of the preferences estimates
				profile = self.interactionsProfileTemplate.generateCopy().reset()
				
				for currPlayer in group:
					preferences = self.playerPrefEstimates[currPlayer]
					for dim in profile.dimensions:
						profile.dimensions[dim] += (preferences.dimensions[dim] / groupSize)

				# print("profile in-configGen: "+str(profile.dimensions)+";groupSize: "+str(groupSize))
				# profile.normalize()
				newConfigProfiles.append(profile)

				# calculate fitness and average state
				currAvgCharacteristics = PlayerCharacteristics()
				currAvgCharacteristics.reset()
				for i in range(len(group)):

					currState = self.playerModelBridge.getPlayerCurrState(group[i])
					currState.profile = profile

					currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
					currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize
				
					if (self.regAlg.isTabular()):
						firstPlayerPreferences = self.playerPrefEstimates[group[i]]
						for j in range(i+1, groupSize):
							currQuality += self.regAlg.predict(firstPlayerPreferences, group[j]) / math.comb(groupSize, 2)

					elif (not self.regAlg.isGroupPredict()):
						currQuality += self.regAlg.predict(profile, group[i])

				if (self.regAlg.isGroupPredict()):
					currQuality += self.regAlg.groupPredict(group)
			
				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(group)
				currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

				newAvgCharacteristics.append(currAvgCharacteristics)
			
			if (currQuality > currMaxQuality):
				bestGroups = newGroups
				bestConfigProfiles = newConfigProfiles
				bestAvgCharacteristics = newAvgCharacteristics
				currMaxQuality = currQuality


			self.completionPerc = i/self.numberOfConfigChoices

		self.updateMetrics(bestGroups)

		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": bestAvgCharacteristics}




class AccuratePRSConfigsGen(ConfigsGenAlg):

	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		simulationFunc, 
		numberOfConfigChoices = None, 
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None, 
		qualityWeights = None):

		super().__init__(
			playerModelBridge = playerModelBridge,
			interactionsProfileTemplate = interactionsProfileTemplate, 
			preferredNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup, 
			minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup, 
			maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup)

		self.simulationFunc = simulationFunc
		self.currIteration = 0

		self.numberOfConfigChoices = 100 if numberOfConfigChoices == None else numberOfConfigChoices 
		self.qualityWeights = PlayerCharacteristics(ability = 0.5, engagement = 0.5) if qualityWeights == None else qualityWeights

	def init(self):
		super().init()
		self.currIteration = 0

	def updateCurrIteration(self, newCurrIteration):
		self.currIteration = newCurrIteration

	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement
	
	def organize(self):
		playerIds = self.playerModelBridge.getAllPlayerIds() 
		minNumGroups = math.ceil(len(playerIds) / self.maxNumberOfPlayersPerGroup)
		maxNumGroups = math.floor(len(playerIds) / self.minNumberOfPlayersPerGroup)

		currMaxQuality = -float("inf")

		bestGroups = []
		bestConfigProfiles = []
		bestAvgCharacteristics = []

		# generate several random groups, calculate their fitness and select the best one
		for i in range(self.numberOfConfigChoices):
			
			# generate several random groups
			newGroups = self.randomConfigGenerator(playerIds, minNumGroups, maxNumGroups)
			newConfigSize = len(newGroups)
			currQuality = 0.0
			newConfigProfiles = []
			newAvgCharacteristics = []

			# generate profiles
			for groupI in range(newConfigSize):
				group = newGroups[groupI]
				groupSize = len(group)

				# generate profile as average of the preferences estimates
				profile = self.interactionsProfileTemplate.generateCopy().reset()
				for currPlayer in group:
					preferences = self.playerModelBridge.getPlayerRealPreferences(currPlayer)
					for dim in profile.dimensions:
						profile.dimensions[dim] += preferences.dimensions[dim] / groupSize
				# profile.normalize()
				newConfigProfiles.append(profile)


				# calculate fitness and average state
				currAvgCharacteristics = PlayerCharacteristics()
				currAvgCharacteristics.reset()
				for currPlayer in group:

					currState = self.playerModelBridge.getPlayerCurrState(currPlayer)
					currState.profile = profile

					currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
					currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize
				
					# does not matter if its executed as bootstrap or not
					newState = self.simulationFunc(
						isBootstrap = False, 
						playerBridge = self.playerModelBridge, 
						state = currState, 
						playerId = currPlayer, 
						currIteration = self.currIteration)
					increases = PlayerState()
					increases.characteristics = PlayerCharacteristics(ability=(newState.characteristics.ability - currState.characteristics.ability), engagement=newState.characteristics.engagement)
					currQuality += self.calcQuality(increases)

				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(group)
				currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

				newAvgCharacteristics.append(currAvgCharacteristics)
			
			if (currQuality > currMaxQuality):
				bestGroups = newGroups
				bestConfigProfiles = newConfigProfiles
				bestAvgCharacteristics = newAvgCharacteristics
				currMaxQuality = currQuality

			self.completionPerc = i/self.numberOfConfigChoices


		self.updateMetrics(bestGroups)

		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": bestAvgCharacteristics}



from deap import base, creator, tools, algorithms
from collections import *

class EvolutionaryConfigsGenDEAP(ConfigsGenAlg):
	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg = None, 
		persEstAlg = None,
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None, 

		initialPopulationSize = None, 
		numberOfEvolutionsPerIteration = None,  
		probOfCross = None, 

		probOfMutation = None,
		probOfMutationConfig = None, 
		probOfMutationGIPs = None, 

		numChildrenPerIteration = None,
		numSurvivors = None,

		cxOp = None,
		
		jointPlayerConstraints = "",
		separatedPlayerConstraints = ""):

		super().__init__(
			playerModelBridge = playerModelBridge,
			interactionsProfileTemplate = interactionsProfileTemplate, 
			preferredNumberOfPlayersPerGroup = preferredNumberOfPlayersPerGroup, 
			minNumberOfPlayersPerGroup = minNumberOfPlayersPerGroup, 
			maxNumberOfPlayersPerGroup = maxNumberOfPlayersPerGroup,
			jointPlayerConstraints = jointPlayerConstraints,
			separatedPlayerConstraints = separatedPlayerConstraints)

		self.regAlg = regAlg
		self.persEstAlg = persEstAlg

		self.initialPopulationSize = 100 if initialPopulationSize == None else initialPopulationSize 


		self.numberOfEvolutionsPerIteration = 500 if numberOfEvolutionsPerIteration == None else numberOfEvolutionsPerIteration
		
		self.probOfMutation = 0.2 if probOfMutation == None else probOfMutation
		self.probOfCross = 0.7 if probOfCross == None else probOfCross
		
		self.probOfMutationConfig = 0.2 if probOfMutationConfig == None else probOfMutationConfig
		self.probOfMutationGIPs = 0.2 if probOfMutationGIPs == None else probOfMutationGIPs
		

		self.numChildrenPerIteration = 5 if numChildrenPerIteration == None else numChildrenPerIteration 
		self.numSurvivors = 5 if numSurvivors == None else numSurvivors 

		if(regAlg==None):
			regAlg = KNNRegression(playerModelBridge, 5)

		self.playerIds = self.playerModelBridge.getAllPlayerIds() 
		self.minNumGroups = math.ceil(len(self.playerIds) / self.maxNumberOfPlayersPerGroup)
		self.maxNumGroups = math.floor(len(self.playerIds) / self.minNumberOfPlayersPerGroup)

		self.searchID = str(id(self)) 

		fitnessFuncId = "FitnessMax"+self.searchID
		individualId = "Individual"+self.searchID

		creator.create(fitnessFuncId, base.Fitness, weights=(1.0,))
		creator.create(individualId, list, fitness=getattr(creator, fitnessFuncId))

		# # conv test
		# creator.create(fitnessFuncId, base.Fitness, weights=(-1.0,))
		# creator.create(individualId, list, fitness=getattr(creator, fitnessFuncId))

		
		self.toolbox = base.Toolbox()

		self.toolbox.register("indices", self.randomIndividualGenerator, self.playerIds, self.minNumGroups, self.maxNumGroups)

		self.toolbox.register("individual", tools.initIterate, getattr(creator, individualId), self.toolbox.indices)
		self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

		self.cxOp = "order" if cxOp == None else cxOp
		if self.cxOp == "order":
			self.toolbox.register("mate", self.cxGIMME_Order)
		else:
			self.toolbox.register("mate", self.cxGIMME_Simple)

		self.toolbox.register("mutate", self.mutGIMME, pGIPs=self.probOfMutationGIPs, pConfig=self.probOfMutationConfig)

		# self.toolbox.register("select", tools.selRoulette)
		# self.toolbox.register("select", tools.selBest, k=self.numFitSurvivors)
		self.toolbox.register("select", self.selGIMME)

		# self.toolbox.register("evaluate", self.calcFitness_convergenceTest)
		self.toolbox.register("evaluate", self.calcFitness)

		self.resetGenAlg()


	def resetGenAlg(self):

		if hasattr(self, "pop"):
			del self.pop
		if hasattr(self, "hof"):
			del self.hof

		self.pop = self.toolbox.population(n = self.initialPopulationSize)
		self.hof = tools.HallOfFame(1)



	def randomIndividualGenerator(self, playerIds, minNumGroups, maxNumGroups):
		groups = self.randomConfigGenerator(playerIds, minNumGroups, maxNumGroups)
		profs = [self.randomProfileGenerator() for i in range(len(groups))]
		return [groups, profs]

	def randomProfileGenerator(self):
		return self.interactionsProfileTemplate.randomized()


	def cxGIMME_Order(self, ind1, ind2):

		# configs
		config1 = ind1[0]
		config2 = ind2[0]

		newConfig1 = []
		newConfig2 = []

		l1 = len(config1)
		l2 = len(config2)

		if (l1 > l2):
			maxLenConfig = config1
			maxLen = l1
			minLenConfig = config2
			minLen = l2
		else:
			maxLenConfig = config2
			maxLen = l2
			minLenConfig = config1
			minLen = l1

		cxpoints = [] 
		
		clist1 = []
		clist2 = []

		remainder1 = []
		remainder2 = []
		for i in range(minLen):
			parent1 = minLenConfig[i]
			parent2 = maxLenConfig[i]

			cxpoint = random.randint(0,len(minLenConfig[i]))
			cxpoints.append(cxpoint) 

			clist1.extend(parent1)
			clist2.extend(parent2)

			remainder1.extend(parent1[cxpoint:])
			remainder2.extend(parent2[cxpoint:])


		d1 = {k:v for v,k in enumerate(clist1)}
		d2 = {k:v for v,k in enumerate(clist2)}

		remainder1.sort(key=d2.get)
		remainder2.sort(key=d1.get)

		for i in range(minLen):
			parent1 = minLenConfig[i]
			parent2 = maxLenConfig[i]

			cxpoint = cxpoints[i] 

			#C1 Implementation
			# maintain left part
			child1, child2 = parent1[:cxpoint], parent2[:cxpoint]


			# reorder right part
			missingLen1 = len(parent1) - len(child1)
			child1.extend(remainder1[:missingLen1])
			remainder1 = remainder1[missingLen1:]

			missingLen2 = len(parent2) - len(child2)
			child2.extend(remainder2[:missingLen2])
			remainder2 = remainder2[missingLen2:]

			newConfig1.append(child1)
			newConfig2.append(child2)


		#the inds become children
		ind1[0] = newConfig1
		ind2[0] = newConfig2

		# breakpoint()

		# profiles are crossed with one point (no need for that when profiles are 1D)
		# breakpoint()
		# if self.interactionsProfileTemplate.dimensionality > 1:
		for i in range(minLen):
			prof1 = ind1[1][i].flattened()
			prof2 = ind2[1][i].flattened()

			newProfiles = tools.cxUniform(prof1, prof2, 0.5)
			# newProfiles = tools.cxOnePoint(prof1, prof2)
			
			#the inds become children
			ind1[1][i] = self.interactionsProfileTemplate.unflattened(newProfiles[0])
			ind2[1][i] = self.interactionsProfileTemplate.unflattened(newProfiles[1])

			# breakpoint()

		del ind1.fitness.values
		del ind2.fitness.values

		return (ind1, ind2)


	def cxGIMME_Simple(self, ind1, ind2):

		# configs
		config1 = ind1[0]
		config2 = ind2[0]

		newConfig1 = []
		newConfig2 = []

		l1 = len(config1)
		l2 = len(config2)

		if (l1 > l2):
			maxLenConfig = config1
			maxLen = l1
			minLenConfig = config2
			minLen = l2
		else:
			maxLenConfig = config2
			maxLen = l2
			minLenConfig = config1
			minLen = l1

		cxpoints = [] 
		
		clist1 = []
		clist2 = []

		remainder1 = []
		remainder2 = []
		for i in range(minLen):
		
			parent1 = [None, None]
			parent2 = [None, None]

			parent1[0] = minLenConfig[i]
			parent1[1] = ind1[1][i].flattened()
			parent2[0] = maxLenConfig[i]
			parent2[1] = ind2[1][i].flattened()

			clist1.append(parent1)
			clist2.append(parent2)


		# breakpoint()

		for ind,clist in zip([ind1,ind2], [clist1,clist2]):


			# print("-----------[Before]-----------")
			# print(json.dumps(ind[1], default=lambda o: o.__dict__))

			randI1 = random.randint(0, len(clist1) - 1)
			randI2 = random.randint(0, len(clist1) - 1)
			

			newProfilesConfig = tools.cxOnePoint(ind1 = clist[randI1][0], ind2 = clist[randI2][0])
			newProfilesGIP = tools.cxOnePoint(ind1 = clist[randI1][1], ind2 = clist[randI2][1])
			
			ind[0][randI1] = newProfilesConfig[0]
			ind[1][randI1] = self.interactionsProfileTemplate.unflattened(newProfilesGIP[0])

			ind[0][randI2] = newProfilesConfig[1]
			ind[1][randI2] = self.interactionsProfileTemplate.unflattened(newProfilesGIP[1])

			# print("-----------[After]-----------")
			# print(json.dumps(ind[1], default=lambda o: o.__dict__))

		# breakpoint()



		del ind1.fitness.values
		del ind2.fitness.values

		return (ind1, ind2)


	def mutGIMME(self, individual, pGIPs, pConfig):
		
		# mutate config
		if random.uniform(0, 1) <= pConfig:
			
			numberOfMutations = 1
			for i in range(numberOfMutations):
				# breakpoint()
				indCpy = copy.copy(individual)
				
				randI1 = random.randint(0, len(indCpy[0]) - 1)
				innerRandI1 = random.randint(0, len(indCpy[0][randI1]) - 1)

				randI2 = innerRandI2 = -1
				while(randI2 < 0 or randI1 == randI2):
					randI2 = random.randint(0, len(indCpy[0]) - 1)
				while(innerRandI2 < 0 or innerRandI1 == innerRandI2):
					innerRandI2 = random.randint(0, len(indCpy[0][randI2]) - 1)


				elem1 = indCpy[0][randI1][innerRandI1]
				elem2 = indCpy[0][randI2][innerRandI2]


				indCpy[0][randI1][innerRandI1] = elem2
				indCpy[0][randI2][innerRandI2] = elem1

				individual[0] = indCpy[0]
				# breakpoint()
			

		#mutate GIPs
		numberOfMutations = 1
		for i in range(numberOfMutations):
			profs = individual[1]
			keys = list(profs[0].dimensions.keys())
			for i in range(len(profs)):
				if random.uniform(0, 1) <= pGIPs:
					# profs[i].randomize()
					for key in keys:
						if random.uniform(0, 1) <= 0.5:
							profs[i].dimensions[key] += random.uniform(0, min(0.2, 1.0 - profs[i].dimensions[key])) 
						else:
							profs[i].dimensions[key] -= random.uniform(0, min(0.2, profs[i].dimensions[key])) 

			individual[1] = profs
		
		del individual.fitness.values
		return individual,


	def reset(self):
		super().reset()
		self.resetGenAlg()

	def calcFitness_convergenceTest(self, individual):
		config = individual[0]
		profiles = individual[1]

		totalFitness = 0.0

		targetConfig = [[0,1,2,3], [4,5,6,7], [8,9,10,11], [12,13,14,15], [16,17,18,19], [20,21,22,23]]

		lenConfig = len(config)
		for groupI in range(lenConfig):
			
			group = config[groupI]
			profile = profiles[groupI]
			
			for playerI in range(len(group)):
				totalFitness += profile.sqrDistanceBetween(InteractionsProfile(dimensions = {'dim_0': 0.98, 'dim_1': 0.005}))
				totalFitness += abs(config[groupI][playerI] - targetConfig[groupI][playerI])
		
		#print(totalFitness)
		totalFitness = totalFitness + 1.0 #helps selection (otherwise Pchoice would always be 0)
		individual.fitness.values = totalFitness,
		return totalFitness, #must return a tuple


	def selGIMME(self, individuals, k, fit_attr="fitness"):
		return tools.selBest(individuals, k, fit_attr)

	def calcFitness(self, individual):
		config = individual[0]
		profiles = individual[1]

		totalFitness = 0.0

		lenConfig = len(config)

		allConstrainsSatisfied = True
		for groupI in range(lenConfig):
			
			group = config[groupI]
			profile = profiles[groupI]


			for constraint in self.jointPlayersConstraints:
				hasToBeInGroup = False
				isNotInGroup = False
				for player in constraint:
					if player in group and isNotInGroup == False:
						hasToBeInGroup = True
					
					elif player not in group and hasToBeInGroup == False:
						isNotInGroup = True

					else:
						allConstrainsSatisfied = False
						break
				
				if allConstrainsSatisfied == False:
					break

			for constraint in self.separatedPlayersConstraints:
				hasToNotBeInGroup = False
				for player in constraint:
					if player in group:
						if hasToNotBeInGroup:
							allConstrainsSatisfied = False
							break
						hasToNotBeInGroup = True
				
				if allConstrainsSatisfied == False:
					break

			for i in range(len(group)):
				if (self.regAlg.isTabular()):
						firstPlayerPreferences = self.playerPrefEstimates[group[i]]
						for j in range(i+1, len(group)):
							totalFitness += self.regAlg.predict(firstPlayerPreferences, group[j]) / math.comb(len(group), 2)

				elif (not self.regAlg.isGroupPredict()):
					totalFitness += self.regAlg.predict(profile, group[i])

			if (self.regAlg.isGroupPredict()):
				totalFitness += self.regAlg.groupPredict(group)
		
		totalFitness = totalFitness + 1.0 #helps selection (otherwise Pchoice would always be 0)
		if allConstrainsSatisfied:
			totalFitness += 1000
		individual.fitness.values = totalFitness,

		return totalFitness, #must return a tuple



	def organize(self):
		self.resetGenAlg()
		if (self.regAlg.isTabular()):
			self.playerPrefEstimates = self.persEstAlg.updateEstimates()
			self.regAlg.playerPrefEstimates = self.playerPrefEstimates

		algorithms.eaMuCommaLambda(
			population = self.pop, 
			toolbox = self.toolbox, 

			lambda_ = self.numChildrenPerIteration, 
			mu = self.numSurvivors, 
			
			cxpb = self.probOfCross, 
			mutpb = self.probOfMutation, 
			
			ngen = self.numberOfEvolutionsPerIteration, 
			
			halloffame = self.hof, 
			verbose = False
		)


		self.completionPerc = len(tools.Logbook())/ self.numberOfEvolutionsPerIteration

		bestGroups = self.hof[0][0]
		bestConfigProfiles = self.hof[0][1]

		# print(bestGroups)
		# print(bestConfigProfiles[0].dimensions)
		# breakpoint()

		avgCharacteristicsArray = []
		for group in bestGroups:
			groupSize = len(group)
			avgCharacteristics = PlayerCharacteristics()
			for currPlayer in group:
				currState = self.playerModelBridge.getPlayerCurrState(currPlayer)
				avgCharacteristics.ability += currState.characteristics.ability / groupSize
				avgCharacteristics.engagement += currState.characteristics.engagement / groupSize
						
				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(group)
				avgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

			avgCharacteristicsArray.append(avgCharacteristics)


		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": avgCharacteristicsArray}

# deterministic algorithms
class ODPIP(ConfigsGenAlg):
	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg,
		persEstAlg,
		taskModelBridge = None,
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None,
		jointPlayerConstraints = "",
		separatedPlayerConstraints = ""):

		super().__init__(playerModelBridge, 
		interactionsProfileTemplate, 
		taskModelBridge,
		preferredNumberOfPlayersPerGroup, 
		minNumberOfPlayersPerGroup, 
		maxNumberOfPlayersPerGroup,
		jointPlayerConstraints = jointPlayerConstraints,
		separatedPlayerConstraints = separatedPlayerConstraints)

		self.regAlg = regAlg
		self.persEstAlg = persEstAlg

		self.coalitionsProfiles = []
		self.coalitionsAvgCharacteristics = []
		self.coalitionsValues = []

		self.playerIds = []	
		

		self.playerPrefEstimates = {}

	

	def getCoalitionInByteFormatValue(self, coalitionInByteFormat):
		coalitionInBitFormat = self.convertCoalitionFromByteToBitFormat(coalitionInByteFormat, len(coalitionInByteFormat))
		return self.f[coalitionInBitFormat]


	def getCoalitionStructureInByteFormatValue(self, coalitionStructure):
		valueOfCS = 0
		for i in range(len(coalitionStructure)):
			valueOfCS += self.getCoalitionInByteFormatValue(coalitionStructure[i])
		
		return valueOfCS


	def convertCoalitionFromByteToBitFormat(self, coalitionInByteFormat, coalitionSize):
		coalitionInBitFormat = 0

		for i in range(coalitionSize):
			coalitionInBitFormat += 1 << (coalitionInByteFormat[i] - 1)

		return coalitionInBitFormat

	# convert group in bit format to group with the player ids
	def getGroupFromBitFormat(self, coalition):
		group = []
		tempCoalition = coalition
		playerNumber = 0
		while tempCoalition != 0:
			if tempCoalition & 1:
				group.append(playerNumber + 1)

			playerNumber += 1
			tempCoalition >>= 1

		
		return group

	def convertFromByteToIds(self, coalition):
		group = []

		for agent in coalition:
			group.append(self.playerIds[agent - 1])

		return group

	def convertFromIdsToBytes(self, coalition):
		group = []

		for agent in coalition:
			for i in range(len(self.playerIds)):
				if self.playerIds[i] == agent:
					group.append(i + 1)

		return group


	def getSizeOfCombinationInBitFormat(self, combinationInBitFormat):
		return bin(combinationInBitFormat).count("1")


	def convertSetOfCombinationsFromBitFormat(self, setOfCombinationsInBitFormat):
		setOfCombinationsInByteFormat = numpy.empty(len(setOfCombinationsInBitFormat), dtype=list)
		for i in range(len(setOfCombinationsInBitFormat)):

			setOfCombinationsInByteFormat[i] = self.getGroupFromBitFormat(setOfCombinationsInBitFormat[i])

		return setOfCombinationsInByteFormat

	def computeAllCoalitionsValues(self):
		numOfAgents = len(self.playerIds)
		numOfCoalitions = 1 << (numOfAgents)

		playersCurrState = {}
		for player in self.playerIds:
			playersCurrState[player] = self.playerModelBridge.getPlayerCurrState(player)


		# (the +- 1 accounts for non divisor cases that need one more/less member)
		adjustedMinSize = self.minNumberOfPlayersPerGroup
		adjustedMaxSize = self.maxNumberOfPlayersPerGroup
		if(adjustedMinSize == adjustedMaxSize and numOfAgents % adjustedMaxSize != 0):
			adjustedMinSize = adjustedMinSize - 1
			adjustedMaxSize = adjustedMaxSize + 1
				
		# initialize all coalitions
		for coalition in range(numOfCoalitions-1, 0, -1):
			group = self.getGroupFromBitFormat(coalition)
			groupInIds = self.convertFromByteToIds(group)

			currQuality = 0.0
			groupSize = len(group)

			# calculate the profile and characteristics only for groups in the range defined 
			if groupSize >= adjustedMinSize and groupSize <= adjustedMaxSize:
				# generate profile as average of the preferences estimates
				profile = self.interactionsProfileTemplate.generateCopy().reset()

				# if (self.regAlg.isTabular()):
				# 	profile = self.findBestProfileForGroup(groupInIds)

				# else:
				for currPlayer in groupInIds:
					preferences = self.playerPrefEstimates[currPlayer]
					for dim in profile.dimensions:
						profile.dimensions[dim] += (preferences.dimensions[dim] / groupSize)

				# calculate fitness and average state
				currAvgCharacteristics = PlayerCharacteristics()
				currAvgCharacteristics.reset()
				for i in range(groupSize):

					currState = playersCurrState[groupInIds[i]]
					currState.profile = profile

					currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
					currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize


					if (self.regAlg.isTabular()):
						firstPlayerPreferences = self.playerPrefEstimates[groupInIds[i]]
						for j in range(i+1, groupSize):
							currQuality += self.regAlg.predict(firstPlayerPreferences, groupInIds[j]) / math.comb(groupSize, 2)

					elif (not self.regAlg.isGroupPredict()):
						currQuality += self.regAlg.predict(profile, groupInIds[i])

				if (self.regAlg.isGroupPredict()):
					currQuality += self.regAlg.groupPredict(groupInIds)
				
				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(groupInIds)
				currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)

				self.coalitionsAvgCharacteristics[coalition] = currAvgCharacteristics
				self.coalitionsProfiles[coalition] = profile
			
			self.coalitionsValues[coalition] = currQuality

	def computeCoalitionsRestrictions(self):
		jointPlayersConstraintInBitFormat = self.jointPlayersConstraints[:]
		separatedPlayersConstraintInBitFormat = self.separatedPlayersConstraints[:]

		for i in range(len(jointPlayersConstraintInBitFormat)):
			jointPlayersConstraintInBitFormat[i] = self.convertFromIdsToBytes(jointPlayersConstraintInBitFormat[i])
			jointPlayersConstraintInBitFormat[i] = self.convertCoalitionFromByteToBitFormat(jointPlayersConstraintInBitFormat[i], len(jointPlayersConstraintInBitFormat[i]))

		for i in range(len(separatedPlayersConstraintInBitFormat)):
			separatedPlayersConstraintInBitFormat[i] = self.convertFromIdsToBytes(separatedPlayersConstraintInBitFormat[i])
			separatedPlayersConstraintInBitFormat[i] = self.convertCoalitionFromByteToBitFormat(separatedPlayersConstraintInBitFormat[i], len(separatedPlayersConstraintInBitFormat[i]))

		return jointPlayersConstraintInBitFormat, separatedPlayersConstraintInBitFormat


	def results(self, cSInByteFormat):
		bestGroups = []
		bestGroupsInBitFormat = []
		bestConfigProfiles = []
		avgCharacteristicsArray = []
	
		#this check was removed because of cases where the preferred group size would not match the integer divisor of the numPlayers
		#for coalition in cSInByteFormat:
			#if not (self.minNumberOfPlayersPerGroup <= len(coalition) <= self.maxNumberOfPlayersPerGroup):
				#return {"groups": [], "profiles": [], "avgCharacteristics": []}
		
		for coalition in cSInByteFormat:
			bestGroups.append(self.convertFromByteToIds(coalition))
			bestGroupsInBitFormat.append(self.convertCoalitionFromByteToBitFormat(coalition, len(coalition)))

		for group in bestGroupsInBitFormat:
			bestConfigProfiles.append(self.coalitionsProfiles[group])
			avgCharacteristicsArray.append(self.coalitionsAvgCharacteristics[group])

		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": avgCharacteristicsArray}

	# function to compute best profile for group according to each players preferences about the task
	def findBestProfileForGroup(self, groupInIds):
		groupSize = len(groupInIds)
		bestProfile = self.interactionsProfileTemplate.generateCopy().reset()

		tasks = self.taskModelBridge.getAllTasksIds()
		
		for playerId in groupInIds:
			bestQuality = -1
			bestTaskId = -1
			
			for taskId in tasks:
				currQuality = self.regAlg.predictTasks(taskId, playerId)

				if currQuality > bestQuality:
					bestQuality = currQuality
					bestTaskId = taskId

			taskProfile = self.taskModelBridge.getTaskInteractionsProfile(bestTaskId)
			for dim in bestProfile.dimensions:
				bestProfile += taskProfile.dimensions[dim] / groupSize

		return bestProfile

	def organize(self):
		self.playerIds = self.playerModelBridge.getAllPlayerIds()
		for i in range(len(self.playerIds)):
			self.playerIds[i] = str(self.playerIds[i])
		self.numPlayers = len(self.playerIds)

		self.coalitionsProfiles = numpy.empty(1 << self.numPlayers, dtype=InteractionsProfile)
		self.coalitionsAvgCharacteristics = numpy.empty(1 << self.numPlayers, dtype=PlayerCharacteristics)
		self.coalitionsValues = numpy.empty(1 << self.numPlayers)

		# estimate preferences
		self.playerPrefEstimates = self.persEstAlg.updateEstimates()

		if (self.regAlg.isTabular()):
			self.regAlg.playerPrefEstimates = self.playerPrefEstimates
			
		# initialization(compute the value for every coalition between min and max number of players)
		self.computeAllCoalitionsValues()
		requiredJointPlayersInBitFormat, restrictedPlayersToJoinInBitFormat = self.computeCoalitionsRestrictions()


		bestCSFound_bitFormat = gs.odpip(self.numPlayers, self.minNumberOfPlayersPerGroup, self.maxNumberOfPlayersPerGroup, self.coalitionsValues.tolist(), requiredJointPlayersInBitFormat, restrictedPlayersToJoinInBitFormat)
		bestCSFound_byteFormat = self.convertSetOfCombinationsFromBitFormat(bestCSFound_bitFormat)

		del bestCSFound_bitFormat

		gc.collect()
		# i = 0
		# for coalition in bestCSFound_byteFormat:
		# 	print("{", end="")
		# 	for player in coalition:
		# 		preferences = self.playerModelBridge.getPlayerRealPreferences(player - 1)
		# 		print(str(player) + ",", end="")

		# 	print(self.coalitionsValues[bestCSFound_bitFormat[i]], end="")
		# 	print("}")

		
		return self.results(bestCSFound_byteFormat)


class CLink(ConfigsGenAlg):
	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg,
		persEstAlg,
		taskModelBridge = None,
		preferredNumberOfPlayersPerGroup = None, 
		minNumberOfPlayersPerGroup = None, 
		maxNumberOfPlayersPerGroup = None):

		super().__init__(playerModelBridge, 
		interactionsProfileTemplate, 
		taskModelBridge,
		preferredNumberOfPlayersPerGroup, 
		minNumberOfPlayersPerGroup, 
		maxNumberOfPlayersPerGroup)

		self.regAlg = regAlg
		self.persEstAlg = persEstAlg

		self.coalitionsProfiles = []
		self.coalitionsAvgCharacteristics = []
		self.coalitionsValues = []

		self.playerIds = []	

		self.playerPrefEstimates = {}


	def getCoalitionInByteFormatValue(self, coalitionInByteFormat):
		coalitionInBitFormat = self.convertCoalitionFromByteToBitFormat(coalitionInByteFormat, len(coalitionInByteFormat))
		return self.f[coalitionInBitFormat]


	def getCoalitionStructureInByteFormatValue(self, coalitionStructure):
		valueOfCS = 0
		for i in range(len(coalitionStructure)):
			valueOfCS += self.getCoalitionInByteFormatValue(coalitionStructure[i])
		
		return valueOfCS


	def convertCoalitionFromByteToBitFormat(self, coalitionInByteFormat, coalitionSize):
		coalitionInBitFormat = 0

		for i in range(coalitionSize):
			coalitionInBitFormat += 1 << (coalitionInByteFormat[i] - 1)

		return coalitionInBitFormat

	# convert group in bit format to group with the player ids
	def getGroupFromBitFormat(self, coalition):
		group = []
		tempCoalition = coalition
		playerNumber = 0
		while tempCoalition != 0:
			if tempCoalition & 1:
				group.append(playerNumber + 1)

			playerNumber += 1
			tempCoalition >>= 1

		
		return group

	def convertFromByteToIds(self, coalition):
		group = []

		for agent in coalition:
			group.append(self.playerIds[agent - 1])

		return group

	def getSizeOfCombinationInBitFormat(self, combinationInBitFormat):
		return bin(combinationInBitFormat).count("1")


	def convertSetOfCombinationsFromBitFormat(self, setOfCombinationsInBitFormat):
		setOfCombinationsInByteFormat = numpy.empty(len(setOfCombinationsInBitFormat), dtype=list)
		for i in range(len(setOfCombinationsInBitFormat)):

			setOfCombinationsInByteFormat[i] = self.getGroupFromBitFormat(setOfCombinationsInBitFormat[i])

		return setOfCombinationsInByteFormat

	def computeAllCoalitionsValues(self):
		numOfAgents = len(self.playerIds)
		numOfCoalitions = 1 << (numOfAgents)

		playersCurrState = {}
		for player in self.playerIds:
			playersCurrState[player] = self.playerModelBridge.getPlayerCurrState(player)

		# initialize all coalitions
		for coalition in range(numOfCoalitions-1, 0, -1):
			group = self.getGroupFromBitFormat(coalition)
			groupInIds = self.convertFromByteToIds(group)

			currQuality = 0.0
			groupSize = len(group)

			# calculate the profile and characteristics only for groups in the range defined
			if groupSize <= self.maxNumberOfPlayersPerGroup:	
				# generate profile as average of the preferences estimates
				profile = self.interactionsProfileTemplate.generateCopy().reset()

				# if (self.regAlg.isTabular()):
				# 	profile = self.findBestProfileForGroup(groupInIds)

				# else:
				for currPlayer in groupInIds:
					preferences = self.playerPrefEstimates[currPlayer]
					for dim in profile.dimensions:
						profile.dimensions[dim] += (preferences.dimensions[dim] / groupSize)

					
				# calculate fitness and average state
				currAvgCharacteristics = PlayerCharacteristics()
				currAvgCharacteristics.reset()
				for i in range(groupSize):

					currState = playersCurrState[groupInIds[i]]
					currState.profile = profile

					currAvgCharacteristics.ability += currState.characteristics.ability / groupSize
					currAvgCharacteristics.engagement += currState.characteristics.engagement / groupSize
				
					if (self.regAlg.isTabular()):
						firstPlayerPreferences = self.playerPrefEstimates[groupInIds[i]]
						for j in range(i+1, groupSize):
							currQuality += self.regAlg.predict(firstPlayerPreferences, groupInIds[j]) / math.comb(groupSize, 2)

					elif (not self.regAlg.isGroupPredict()):
						currQuality += self.regAlg.predict(profile, groupInIds[i])

				if (self.regAlg.isGroupPredict()):
					currQuality += self.regAlg.groupPredict(groupInIds)
			
				diversityValueAlg = DiversityValueAlg(self.playerModelBridge, 0)
				personalities = diversityValueAlg.getPersonalitiesListFromPlayerIds(groupInIds)
				currAvgCharacteristics.group_diversity = diversityValueAlg.getTeamPersonalityDiveristy(personalities)
						
				self.coalitionsAvgCharacteristics[coalition] = currAvgCharacteristics
				self.coalitionsProfiles[coalition] = profile
			
			self.coalitionsValues[coalition] = currQuality


	def results(self, cSInByteFormat):
		bestGroups = []
		bestGroupsInBitFormat = []
		bestConfigProfiles = []
		avgCharacteristicsArray = []
		for coalition in cSInByteFormat:
			bestGroups.append(self.convertFromByteToIds(coalition))
			bestGroupsInBitFormat.append(self.convertCoalitionFromByteToBitFormat(coalition, len(coalition)))

		for group in bestGroupsInBitFormat:
			bestConfigProfiles.append(self.coalitionsProfiles[group])
			avgCharacteristicsArray.append(self.coalitionsAvgCharacteristics[group])

		return {"groups": bestGroups, "profiles": bestConfigProfiles, "avgCharacteristics": avgCharacteristicsArray}

	# function to compute best profile for group according to each players preferences about the task
	def findBestProfileForGroup(self, groupInIds):
		groupSize = len(groupInIds)
		bestProfile = self.interactionsProfileTemplate.generateCopy().reset()

		tasks = self.taskModelBridge.getAllTasksIds()
		
		for playerId in groupInIds:
			bestQuality = -1
			bestTaskId = -1
			
			for taskId in tasks:
				currQuality = self.regAlg.predictTasks(taskId, playerId)

				if currQuality > bestQuality:
					bestQuality = currQuality
					bestTaskId = taskId

			taskProfile = self.taskModelBridge.getTaskInteractionsProfile(bestTaskId)
			for dim in bestProfile.dimensions:
				bestProfile += taskProfile.dimensions[dim] / groupSize

		return bestProfile

	def organize(self):
		self.playerIds = self.playerModelBridge.getAllPlayerIds()
		for i in range(len(self.playerIds)):
			self.playerIds[i] = str(self.playerIds[i])
		self.numPlayers = len(self.playerIds)

		self.coalitionsProfiles = numpy.empty(1 << self.numPlayers, dtype=InteractionsProfile)
		self.coalitionsAvgCharacteristics = numpy.empty(1 << self.numPlayers, dtype=PlayerCharacteristics)
		self.coalitionsValues = numpy.empty(1 << self.numPlayers)

		# estimate preferences
		self.playerPrefEstimates = self.persEstAlg.updateEstimates()

		if (self.regAlg.isTabular()):
			self.regAlg.playerPrefEstimates = self.playerPrefEstimates

		# initialization(compute the value for every coalition between min and max number of players)
		self.computeAllCoalitionsValues()

		bestCSFound_bitFormat = (gs.clink(self.numPlayers, self.minNumberOfPlayersPerGroup, self.maxNumberOfPlayersPerGroup, self.coalitionsValues.tolist()))
		bestCSFound_byteFormat = self.convertSetOfCombinationsFromBitFormat(bestCSFound_bitFormat)

		del bestCSFound_bitFormat
		
		gc.collect()


		return self.results(bestCSFound_byteFormat)



from setuptools import dist, setup, Extension
import glob

sfc_module = Extension(
  name="GIMMESolver",
  include_dirs = ['solverModules/GIMME_solver_modules'],
  sources = glob.glob('solverModules/GIMME_solver_modules/*.cpp'),
  language='c++',
  )

with open('README.md', 'r') as file:
    readmeFile = file.read()

setup(
    name="GIMMECore",
    version="1.6.5",
    license = "CC BY 4.0",
    python_requires='>=3.7',
    author="Samuel Gomes",
    author_email = "samuel.gomes@tecnico.ulisboa.pt",
    description="GIMME (Group Interactions Management for Multiplayer sErious games) is a research tool which focuses on the management of interactions in groups so that the collective ability improves.",
    long_description=readmeFile,
    long_description_content_type="text/markdown",
    url='https://github.com/SamGomes/GIMME',
    packages=['GIMMECore', 'GIMMECore.ModelBridge', 'GIMMECore.AlgDefStructs'],
    classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    install_requires=[
      'python-decouple>=3.0',
      'deap>=1.3.3',
      'pandas>=1.3.5',
      'scikit-learn>=1.0.2',
      'matplotlib>=3.5.3',
      'pymongo>=4.3.3'
    ],
    ext_modules=[sfc_module]
)




from abc import ABC, abstractmethod
from ..PlayerStructs import *
import json

class PreferencesEstAlg(ABC):

	def __init__(self, playerModelBridge):
		self.playerModelBridge = playerModelBridge

	@abstractmethod
	def updateEstimates(self):
		pass



class ExploitationPreferencesEstAlg(PreferencesEstAlg):
	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg, 
		qualityWeights = None):

		super().__init__(playerModelBridge)
		
		self.playerModelBridge = playerModelBridge
		self.qualityWeights = PlayerCharacteristics(ability = 0.5, engagement = 0.5) if qualityWeights == None else qualityWeights 

		self.interactionsProfileTemplate = interactionsProfileTemplate
		self.regAlg = regAlg
		self.bestQualities = {}

	def calcQuality(self, state):
		return self.qualityWeights.ability*state.characteristics.ability + self.qualityWeights.engagement*state.characteristics.engagement
	
	
	def updateEstimates(self):
		playerIds = self.playerModelBridge.getAllPlayerIds()
		for playerId in playerIds:
			currPreferencesEst = self.playerModelBridge.getPlayerPreferencesEst(playerId)
			currPreferencesQuality = self.bestQualities.get(playerId, 0.0)
			lastDataPoint = self.playerModelBridge.getPlayerCurrState(playerId)
			quality = self.calcQuality(lastDataPoint)
			if quality > currPreferencesQuality:
				self.bestQualities[playerId] = currPreferencesQuality
				self.playerModelBridge.setPlayerPreferencesEst(playerId, lastDataPoint.profile)




class ExplorationPreferencesEstAlg(PreferencesEstAlg):
	def __init__(self, 
		playerModelBridge, 
		interactionsProfileTemplate, 
		regAlg, 
		numTestedPlayerProfiles = None):
		
		super().__init__(playerModelBridge)
		
		self.playerModelBridge = playerModelBridge

		self.numTestedPlayerProfiles = 100 if numTestedPlayerProfiles == None else numTestedPlayerProfiles
		self.interactionsProfileTemplate = interactionsProfileTemplate

		self.regAlg = regAlg


	def updateEstimates(self):
		playerIds = self.playerModelBridge.getAllPlayerIds()
		updatedEstimates = {}
		for playerId in playerIds:
			
			currPreferencesEst = self.playerModelBridge.getPlayerPreferencesEst(playerId)
			newPreferencesEst = currPreferencesEst
			if(currPreferencesEst != None):
				bestQuality = self.regAlg.predict(currPreferencesEst, playerId)
			else:
				bestQuality = -1
			
			for i in range(self.numTestedPlayerProfiles):
				profile = self.interactionsProfileTemplate.generateCopy().randomize()
				currQuality = self.regAlg.predict(profile, playerId)
				if currQuality >= bestQuality:
					bestQuality = currQuality
					newPreferencesEst = profile

			self.playerModelBridge.setPlayerPreferencesEst(playerId, newPreferencesEst)
			updatedEstimates[str(playerId)] = newPreferencesEst


		
		return updatedEstimates
